use std::path::{Path, PathBuf};
// use std::io::BufReader;
use std::thread;
use std::time::Duration;
use std::fs::File;
use std::io::{Cursor, Read};
use std::sync::Arc;
use rodio::Decoder;
use crate::rodio::Source;

use std::io::BufWriter;
use std::io::BufReader;
use std::collections::HashMap;

// pub fn collect



#[derive(Serialize, Deserialize, Clone, Default, Debug)]
pub struct Sound {
    pub name: String,
    pub location: PathBuf,
    #[serde(skip)]
    data: Arc<Vec<u8>>,
    pub volume: f32,
    pub roll: i32,
    pub rollrate: i32,
    pub trim: f32,
    pub active: bool,
    pub speed: f32
}


impl AsRef<[u8]> for Sound {
    fn as_ref(&self) -> &[u8] {
        &self.data
    }
}

impl Sound {
    pub fn new(filename: &str) -> std::io::Result<Sound> {
        let path = PathBuf::from(filename);
        let mut buf = Vec::new();
        let mut file = File::open(filename)?;
        file.read_to_end(&mut buf)?;
        Ok(Sound{
            data: Arc::new(buf),
            name: path.file_name().unwrap_or_default().to_string_lossy().to_string(),
            location: path,
            volume: 1.0,
            speed: 1.0,
            ..Default::default()
        })
    }

    pub fn reload(mut self) {
        if let Ok(s) = Sound::new(&self.location.to_string_lossy()) {
            self.data = s.data;
        }
    }

    pub fn reload_data(&self) -> Result<Sound, std::io::Error>{
        Sound::new(&self.location.to_string_lossy())
    }

    pub fn cursor(self: &Self) -> Cursor<Sound> {
        Cursor::new(
            Sound {
                data: self.data.clone(),
                ..Default::default()
            })
    }

    pub fn decoder(self: &Self) -> Decoder<Cursor<Sound>> {
        Decoder::new(self.cursor()).unwrap()
    }

    pub fn play(&self, device: &rodio::Device) {
        // let s = self.decoder().convert_samples()
        rodio::play_raw(&device, self.decoder()
        .speed(self.speed)
        .amplify(self.volume)
        .convert_samples());
        // self.active = true;

    }
}

#[derive(Serialize, Deserialize, Clone, Debug)]
pub struct Pattern {
    pub name: String,
    pub repeat: i32,
    pub bpm: i32,
    pub xsize: usize,
    pub ysize: usize,
    pub sounds: HashMap<(usize, usize), Sound>,
}

impl Default for Pattern {
    fn default() -> Pattern {
        Pattern {
            name: "Unnamed".to_string(),
            repeat: 0,
            bpm: 120,
            xsize: 8,
            ysize: 1,
            sounds: HashMap::new()
        }
    }
}

impl Pattern {
    pub fn play(&self, device: &rodio::Device) {
        let delay = ((60.0 / self.bpm as f32)*1000.0) as u64;
        // loop for repeats
        for _ in 0..self.repeat + 1 {
            for x in 0..self.xsize {
                for y in 0..self.ysize {
                    if let Some(s) = self.sounds.get(&(x, y)) {
                        s.play(device);
                    }
                }
                thread::sleep(Duration::from_millis(delay));
            }
        }
    }

    pub fn load(file: &Path) -> Result<Pattern, String> {
        match File::open(file) {
            Ok(f) => {
                serde_json::from_reader(BufReader::new(f))
                    .map_err(|e| e.to_string())
                    // .map(|p: Pattern| p.reload_sounds( ))
            }
            Err(e) => Err(e.to_string())
        }
    }
    pub fn replace_sounds(&mut self, new_sound: &Sound, row: usize) {
        for x in 0..self.xsize {
            if let Some(snd) = self.sounds.get_mut(&(x, row)) {
                snd.location = new_sound.location.clone();
                snd.data = new_sound.data.clone();
            }
        
        }

    }
    pub fn clear_sounds(&mut self, row: usize) {
        for x in 0..self.xsize {
            if let Some(_) = self.sounds.get(&(x, row)) {
                self.sounds.remove(&(x, row));
            }
        }
    }

    pub fn reload_sounds(&mut self) -> Pattern {
        let mut newp = self.clone();
        for x in 0..self.xsize {
            for y in 0..self.ysize {
                if let Some(snd) = self.sounds.get_mut(&(x,y)) {
                    // dbg!(&snd);
                    snd.reload_data().unwrap();
                    // let s = snd.reload_data().unwrap();
                    // snd.data = s.data.clone();
                }
            }
        }
        newp.clone()
    }

    pub fn save(&self) {
        let pattern_dir = Path::new("patterns");

        let writer = BufWriter::new(File::create(pattern_dir.join(&self.name).with_extension("pat")).unwrap());
        serde_json::to_writer_pretty(writer, &self).unwrap();
    }

    pub fn extend_length(&mut self, slots: usize) {
        self.xsize = self.xsize+slots;
    }

    pub fn extend_row(&mut self, rows: usize) {
        self.ysize = self.ysize+rows;
    }


}

#[derive(Clone, Debug)]
pub struct Timeline {
    pub patterns: Vec<Pattern>
}

impl Default for Timeline {
    fn default() -> Timeline {
        Timeline {
            patterns: vec![Pattern::default()]
        }
    }
}